# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Model that wraps all segmentation models of mmdetection
"""
import logging
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import torch.nn
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.segmentation import PolygonType, Segmentation
from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.api.model import SegmentationModel
from mlcvzoo_base.api.structs import Runtime
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from mmdet.apis.det_inferencer import DetInferencer
from mmdet.evaluation import INSTANCE_OFFSET
from mmdet.structures.det_data_sample import DetDataSample
from mmdet.structures.mask.structures import bitmap_to_polygon
from nptyping import Int, NDArray, Shape

from mlcvzoo_mmdetection.configuration import (
    MMDetectionConfig,
    MMDetectionInferenceConfig,
)
from mlcvzoo_mmdetection.model import MMDetectionModel

try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.inferencer import MMDeployInferencer
except ModuleNotFoundError as error:
    MMDeployInferencer = None  # type: ignore # pylint: disable=invalid-name

logger = logging.getLogger(__name__)

ImageType = NDArray[Shape["Height, Width, Any"], Int]


class MMSegmentationModel(
    MMDetectionModel[MMDetectionInferenceConfig],
    SegmentationModel[MMDetectionConfig, Union[str, ImageType]],
    NetBased[torch.nn.Module, MMDetectionInferenceConfig],
    Trainable,
):
    """Wrapper for MMDetection segmentation models."""

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
        is_multi_gpu_instance: bool = False,
        runtime: str = Runtime.DEFAULT,
    ) -> None:
        MMDetectionModel.__init__(
            self,
            configuration=self.create_configuration(
                from_yaml, configuration, string_replacement_map
            ),
            string_replacement_map=string_replacement_map,
            init_for_inference=init_for_inference,
            is_multi_gpu_instance=is_multi_gpu_instance,
            runtime=runtime,
        )
        SegmentationModel.__init__(
            self,
            configuration=self.configuration,
            mapper=AnnotationClassMapper(
                class_mapping=self.configuration.class_mapping,
                reduction_mapping=self.configuration.inference_config.reduction_class_mapping,
            ),
            init_for_inference=init_for_inference,
            runtime=runtime,
        )
        NetBased.__init__(self, net=self.net)
        Trainable.__init__(self)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> MMDetectionConfig:
        return cast(
            MMDetectionConfig,
            create_basis_configuration(
                configuration_class=MMDetectionConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @property
    def num_classes(self) -> int:
        return self.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.mapper.annotation_class_id_to_model_class_name_map

    def __decode_mmdet_result(self, prediction: Any) -> List[Segmentation]:
        """
        Decode output of an object detection model from mmdetection

        Args:
            prediction: The result that the model has predicted

        Returns:
            The prediction as list of bounding boxes in MLCVZoo format
        """

        segmentations: List[Segmentation] = []

        if hasattr(prediction, "pred_panoptic_seg"):
            # Have a shape of (IMAGE_HEIGHT, IMAGE_WIDTH) with each pixel
            # stating the class of the segmentation
            pan_results = prediction.pred_panoptic_seg.sem_seg.cpu().numpy()[0]
            # Get the classes that are contained in this panoptic segmentation
            valid_indices = np.unique(pan_results)[::-1]
            # Since the panoptic result has an INSTANCE_OFFSET, we need to make
            # sure not to get the class-id which is equal to the number of classes
            legal_indices = valid_indices != self.num_classes
            valid_indices = valid_indices[legal_indices]
            pan_class_ids = np.array(  # type: ignore[var-annotated]
                [class_id % INSTANCE_OFFSET for class_id in valid_indices],
                dtype=np.int64,
            )
            # Using [None] index to match the shapes:
            # pan_results (H, W) => (1, H, W)
            # valid_indices (N) => (N, 1, 1)
            # panoptic_segmentations => (N, H, W) each dimension corresponds to a boolean mask
            #                                     for the specific 'valid' class
            panoptic_segmentations = pan_results[None] == valid_indices[:, None, None]

            for panoptic_segmentation, class_id in zip(
                panoptic_segmentations, pan_class_ids
            ):
                score = 1.0

                contours, with_hole = bitmap_to_polygon(panoptic_segmentation)

                # Need to check with len operator due to numpy list
                if len(contours) == 0:
                    continue

                segmentations.append(
                    Segmentation(
                        class_identifier=ClassIdentifier(
                            class_id=class_id,
                            class_name=self.mapper.map_annotation_class_id_to_model_class_name(
                                class_id=class_id
                            ),
                        ),
                        score=score,
                        # For now, we only take the first contour and don't expect
                        # segmentations to be split into multiple parts
                        polygon=cast(PolygonType, contours[0]),
                        difficult=False,
                        occluded=False,
                        content="",
                    )
                )
        else:
            valid_indices = (
                prediction.pred_instances.scores
                > self.configuration.inference_config.score_threshold
            )

            # Filter results according to the determined valid indices
            valid_masks = prediction.pred_instances.masks[valid_indices]
            valid_class_ids = prediction.pred_instances.labels[valid_indices]
            valid_scores = prediction.pred_instances.scores[valid_indices]

            for i, (mask, class_id, score) in enumerate(
                zip(valid_masks, valid_class_ids, valid_scores)
            ):
                np_mask = np.stack(mask.detach().cpu().numpy(), axis=0)
                contours, with_hole = bitmap_to_polygon(np_mask)

                # Need to check with len operator due to numpy list
                if len(contours) == 0:
                    continue

                class_id_int = int(class_id.cpu())

                # TODO: Add the 'box' parameter to build_segmentations to simplify the
                #       instantiation of the segmentations per class_identifier
                class_identifiers = (
                    self.mapper.map_model_class_id_to_output_class_identifier(
                        class_id=class_id_int
                    )
                )

                model_class_identifier = ClassIdentifier(
                    class_id=class_id_int,
                    class_name=self.mapper.map_annotation_class_id_to_model_class_name(
                        class_id=class_id_int
                    ),
                )

                for class_identifier in class_identifiers:
                    segmentations.append(
                        Segmentation(
                            class_identifier=class_identifier,
                            model_class_identifier=model_class_identifier,
                            score=float(score),
                            # For now, we only take the first contour and don't expect
                            # segmentations to be split into multiple parts
                            polygon=contours[0],
                            difficult=False,
                            occluded=False,
                            content="",
                        )
                    )

        return segmentations

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[Segmentation]]:
        """Run a prediction on a single input.

        Args:
            data_item (Union[str, ImageType]): The input to run the inference on.

        Raises:
            ValueError: If the net attribute is not initialized and the runtime is 'DEFAULT'.
            ValueError: If the inferencer aatribute is not initialized.
            RuntimeError: If the model is deployed with MMDeploy and the mmdeploy module can not be
                imported.
            ValueError: If the runtime is not supported.

        Returns:
            Tuple[Union[str, ImageType], List[Segmentation]]: The input and the predicted
                segmentations.
        """

        no_inferencer_error = ValueError(
            "The 'inferencer' attribute is not initialized, make sure to instantiate with "
            "init_for_inference=True"
        )

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )
            if self.inferencer is None:
                raise no_inferencer_error

            return data_item, self.__decode_mmdet_result(
                prediction=cast(DetInferencer, self.inferencer)(
                    inputs=data_item, return_datasample=True, batch_size=1
                )["predictions"][0]
            )
        if self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.inferencer is None:
                raise no_inferencer_error

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            return data_item, self.__decode_mmdet_result(
                prediction=cast(MMDeployInferencer, self.inferencer)(data_item)[0]
            )

        raise ValueError(f"Prediction is not supported for runtime '{self.runtime}'.")

    def predict_many(
        self, data_items: List[Union[str, ImageType]]
    ) -> List[Tuple[Union[str, ImageType], List[Segmentation]]]:
        """Run a prediction on a batch of input.

        Args:
            data_items (List[Union[str, ImageType]]): The inputs to run the inference on.

        Raises:
            ValueError: If the net attribute is not initialized and the runtime is 'DEFAULT'.
            ValueError: If the inferencer aatribute is not initialized.
            RuntimeError: If the model is deployed with MMDeploy and the mmdeploy module can not be
                imported.
            ValueError: If the runtime is not supported.

        Returns:
            List[Tuple[Union[str, ImageType], List[Segmentation]]]: A list of inputs and the
                predicted segmentations.
        """

        no_inferencer_error = ValueError(
            "The 'inferencer' attribute is not initialized, make sure to instantiate with "
            "init_for_inference=True"
        )

        prediction_list: List[Tuple[Union[str, ImageType], List[Segmentation]]] = []

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )
            if self.inferencer is None:
                raise no_inferencer_error

            # TODO: add batch-size as parameter
            predictions: List[DetDataSample] = cast(DetInferencer, self.inferencer)(
                inputs=data_items,
                return_datasample=True,
                batch_size=len(data_items),
            )["predictions"]
        elif self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.inferencer is None:
                raise no_inferencer_error

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            predictions = [
                cast(MMDeployInferencer, self.inferencer)(data_item)[0]
                for data_item in data_items
            ]
        else:
            raise ValueError(
                f"Multi-prediction is not supported for runtime '{self.runtime}'."
            )

        for data_item, prediction in zip(data_items, predictions):
            segmentations = self.__decode_mmdet_result(prediction=prediction)

            prediction_list.append(
                (
                    data_item,
                    segmentations,
                )
            )

        return prediction_list


if __name__ == "__main__":
    MMDetectionModel.run(MMSegmentationModel)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for providing the possibility to train a mmdetection
model on data that is provided by the annotation handler
of the MLCVZoo. This is realized by extending the 'DATASETS'
registry of mmdetection.
"""

import logging
import os
from typing import Any, Dict, List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.data_preparation.annotation_handler import AnnotationHandler
from mmdet.datasets.base_det_dataset import BaseDetDataset
from mmengine.config import ConfigDict
from related import to_model

logger = logging.getLogger(__name__)


class MLCVZooMMDetDataset(BaseDetDataset):
    """
    Implementation of a custom dataset. It follows the instructions given by:
    https://mmdetection.readthedocs.io/en/dev-3.x/advanced_guides/customize_dataset.html

    We followed an example and created our own dataset class
    which has to be compatible to the class "BaseDataset"
    of the mmdetection framework.
    """

    def __init__(
        self,
        *args: Any,
        annotation_handler_config: Optional[ConfigDict] = None,
        class_mapping_config: Optional[ConfigDict] = None,
        include_segmentation_as_bbox: bool = False,
        **kwargs: Any,
    ) -> None:
        self.annotations: List[BaseAnnotation] = []
        self.CLASSES: List[str] = []

        self._include_segmentation_as_bbox = include_segmentation_as_bbox

        if annotation_handler_config is not None:
            # Set annotation handler configuration from given annotation handler config
            configuration = to_model(AnnotationHandlerConfig, annotation_handler_config)
            # Set class mapping configuration from given class mapping config
            if class_mapping_config is not None:
                configuration.class_mapping = to_model(
                    ClassMappingConfig, class_mapping_config
                )
            # Create annotation handler, if no class mapping config was given,
            # assume it is part of the config file
            annotation_handler = AnnotationHandler(
                configuration=configuration,
            )
            # Load annotations and model classes from annotation handler
            self.annotations = annotation_handler.parse_training_annotations()
            self.CLASSES = annotation_handler.mapper.get_model_class_names()

            self.METAINFO = {
                "classes": self.CLASSES,
            }

        self._with_bbox: bool = False
        self._with_mask: bool = False
        self._box_type: str = "hbox"
        pipeline = kwargs.get("pipeline")
        if pipeline is not None:
            annotation_pipelines = []
            for pipeline_element in pipeline:
                if pipeline_element["type"] in self._get_annotation_pipeline_types():
                    annotation_pipelines.append(pipeline_element)

            # Handle with_seg if we need to train panoptic segmentation models
            # self.with_seg: bool
            if len(annotation_pipelines) > 0:
                self._with_bbox = annotation_pipelines[0].get("with_bbox", False)
                self._with_mask = annotation_pipelines[0].get(
                    "with_mask", False
                ) or annotation_pipelines[0].get("with_polygon", False)
                self._box_type = annotation_pipelines[0].get("box_type", "hbox")

        BaseDetDataset.__init__(self, *args, **kwargs)

    @staticmethod
    def _get_annotation_pipeline_types() -> List[str]:
        """
        List of types which are associated with pipelines that load annotations.
        These are used to gather information which kind of data e.g. bounding boxes or
        segmentations should be parsed in the load_data_list(...).

        Returns:
            List of types as strings
        """
        return ["LoadAnnotations", "mmdet.LoadAnnotations"]

    @staticmethod
    def __generate_summary_message(base_message: str, data_dict: Dict[str, int]) -> str:
        """
        Produce a data summary message

        Args:
            base_message: Header of the message
            data_dict: Dictionary containing the summary information

        Returns:
            The generated message
        """
        message = base_message.format(sum(data_dict.values())) + "\n"

        for key, value in data_dict.items():
            message += f"{str(value).rjust(10)} | {str(key)}\n"

        return message

    def load_data_list(self) -> List[Dict[str, Any]]:
        """
        Transform the mlcvzoo annotations into the format that is required by the data
        loaders of the open-mmlab world.

        For the respective annotation format convention have a look at their documentation

        Returns:
             The loaded data list in the format required by the data loader in the open-mmlab world
        """
        data_list: List[Dict[str, Any]] = []
        bbox_data_dict: Dict[str, int] = {}
        segmentation_data_dict: Dict[str, int] = {}

        # TODO: Handle ignore flag if present in kwargs
        for img_id, annotation in enumerate(self.annotations):
            if not os.path.isfile(annotation.image_path):
                logger.debug(
                    "Skip annotation with path='%s' since image='%s' does not exist",
                    annotation.annotation_path,
                    annotation.image_path,
                )
                continue

            instances: List[Dict[str, Any]] = []
            if self._with_bbox and not self._with_mask:
                for bounding_box in annotation.get_bounding_boxes(
                    include_segmentations=self._include_segmentation_as_bbox
                ):
                    # Check which box type is configured for the dataset pipeline

                    # Object Detection training - only take orthogonal boxes
                    # hbox => xyxy
                    if self._box_type == "hbox":
                        mmdet_bbox = bounding_box.ortho_box().to_list(dst_type=float)

                    # Rotated Object Detection training - every box is supported
                    # qbox => xyxyxyxy
                    elif self._box_type == "qbox":
                        mmdet_bbox = bounding_box.polygon().reshape(-1).tolist()
                    else:
                        raise ValueError(
                            "Currently only 'hbox' and 'qbox' are supported "
                            "for the parameter 'box_type'"
                        )

                    instances.append(
                        {
                            "ignore_flag": 0,
                            "ignore": 0,
                            "bbox_label": bounding_box.class_id,
                            "bbox": mmdet_bbox,
                        }
                    )
                    if bounding_box.class_identifier.class_name not in bbox_data_dict:
                        bbox_data_dict[bounding_box.class_identifier.class_name] = 1
                    else:
                        bbox_data_dict[bounding_box.class_identifier.class_name] += 1
            elif self._with_mask:
                # Segmentation training - train with masks and bounding boxes
                for segmentation in annotation.segmentations:
                    instances.append(
                        {
                            "ignore_flag": 0,
                            "ignore": 0,
                            "bbox_label": segmentation.class_id,
                            "bbox": segmentation.ortho_box().to_list(dst_type=float),
                            # Mask annotations for usage in Instance/Panoptic Segmentation models.
                            # mmdet allows two formats: list[list[float]] or dict,
                            # we use the list format here, where the inner list[float] has to be
                            # in the format [x1, y1, ..., xn, yn] (n≥3)
                            "mask": segmentation.polygon().reshape(1, -1).tolist(),
                            # Polygon annotations for training Text Detection networks in mmocr.
                            # The format is the same as above for the inner list:
                            # [x1, y1, ..., xn, yn] (n≥3)
                            "polygon": segmentation.polygon().reshape(-1).tolist(),
                        }
                    )
                    if (
                        segmentation.class_identifier.class_name
                        not in segmentation_data_dict
                    ):
                        segmentation_data_dict[
                            segmentation.class_identifier.class_name
                        ] = 1
                    else:
                        segmentation_data_dict[
                            segmentation.class_identifier.class_name
                        ] += 1

            data_list.append(
                {
                    "img_path": annotation.image_path,
                    "img_id": img_id,
                    "seg_map_path": None,
                    "height": annotation.get_height(),
                    "width": annotation.get_width(),
                    "instances": instances,
                }
            )

        if len(bbox_data_dict.values()) > 0:
            logger.info(
                MLCVZooMMDetDataset.__generate_summary_message(
                    base_message="MLCVZooMMDetDataset loaded {} BBoxes:",
                    data_dict=bbox_data_dict,
                )
            )

        if len(segmentation_data_dict.values()) > 0:
            logger.info(
                MLCVZooMMDetDataset.__generate_summary_message(
                    base_message="MLCVZooMMDetDataset loaded {} Segmentations: ",
                    data_dict=segmentation_data_dict,
                )
            )

        return data_list



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Model that wraps all objection detection models of mmdetection
"""
from __future__ import annotations

import argparse
import inspect
import logging
import os
import shlex
import subprocess
import sys
from abc import ABC
from threading import Thread
from typing import IO, Any, Callable, Dict, Optional, TypeVar, Union, cast

import torch
import torch.nn
import yaml
from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.api.model import Model
from mlcvzoo_base.api.structs import Runtime
from mmdet.apis.det_inferencer import DetInferencer
from mmdet.registry import DATASETS
from mmdet.utils import setup_cache_size_limit_of_dynamo
from mmengine.config import Config
from mmengine.runner import Runner, load_checkpoint
from nptyping import Int, NDArray, Shape

from mlcvzoo_mmdetection.configuration import (
    MMDetectionConfig,
    MMDetectionInferenceConfig,
    MMDetectionMMDeployConfig,
)
from mlcvzoo_mmdetection.mlcvzoo_mmdet_dataset import MLCVZooMMDetDataset
from mlcvzoo_mmdetection.utils import init_mm_config, modify_config

try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.inferencer import MMDeployInferencer
except ModuleNotFoundError:
    MMDeployInferencer = None  # type: ignore # pylint: disable=invalid-name
try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.converter import MMDeployConverter
except ModuleNotFoundError:
    MMDeployConverter = None  # type: ignore # pylint: disable=invalid-name

logger = logging.getLogger(__name__)

ImageType = NDArray[Shape["Height, Width, Any"], Int]

MMDetectionInferenceConfigType = TypeVar(  # pylint: disable=invalid-name
    "MMDetectionInferenceConfigType", bound=MMDetectionInferenceConfig
)


class MMDetectionModel(
    Model[Any, Any, Any],
    NetBased[torch.nn.Module, MMDetectionInferenceConfigType],
    Trainable,
    ABC,
):
    """Wrapper for MMDetection models."""

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
        is_multi_gpu_instance: bool = False,
        runtime: str = Runtime.DEFAULT,
    ) -> None:
        self.net: Optional[torch.nn.Module] = None
        self.inferencer: Optional[Union[DetInferencer, MMDeployInferencer]] = None

        self.yaml_config_path: Optional[str] = from_yaml
        self.is_multi_gpu_instance: bool = is_multi_gpu_instance

        self.configuration: MMDetectionConfig = self.create_configuration(
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        self.configuration.recursive_string_replacement(
            string_replacement_map=string_replacement_map
        )
        self._register_dataset()

        self.__update_mm_config(init_for_inference=init_for_inference)

        # config path to mmdetection specific .py file is set
        if self.configuration.mm_config.config_path:
            if string_replacement_map:
                self.configuration.mm_config.config_path = modify_config(
                    config_path=self.configuration.mm_config.config_path,
                    string_replacement_map=string_replacement_map,
                )

        # config path to MMDeploy specific .py file for ONNX Runtime is set
        if (
            self.configuration.mmdeploy_onnxruntime_config
            and self.configuration.mmdeploy_onnxruntime_config.config_path
        ) and string_replacement_map:
            self.configuration.mmdeploy_onnxruntime_config.config_path = modify_config(
                config_path=self.configuration.mmdeploy_onnxruntime_config.config_path,
                string_replacement_map=string_replacement_map,
            )

        # config path to MMDeploy specific .py file for ONNX Runtime for float16 precision is set
        if (
            self.configuration.mmdeploy_onnxruntime_float16_config
            and self.configuration.mmdeploy_onnxruntime_float16_config.config_path
        ) and string_replacement_map:
            self.configuration.mmdeploy_onnxruntime_float16_config.config_path = modify_config(
                config_path=self.configuration.mmdeploy_onnxruntime_float16_config.config_path,
                string_replacement_map=string_replacement_map,
            )

        # config path to MMDeploy specific .py file for TensorRT is set
        if (
            self.configuration.mmdeploy_tensorrt_config
            and self.configuration.mmdeploy_tensorrt_config.config_path
        ) and string_replacement_map:
            self.configuration.mmdeploy_tensorrt_config.config_path = modify_config(
                config_path=self.configuration.mmdeploy_tensorrt_config.config_path,
                string_replacement_map=string_replacement_map,
            )

        self.cfg = init_mm_config(
            mm_config=self.configuration.mm_config,
        )
        self.__init_dataloader()
        self.__deploy(runtime=runtime, init_for_inference=init_for_inference)

        Model.__init__(
            self,
            configuration=self.configuration,
            init_for_inference=init_for_inference,
            runtime=runtime,
        )
        NetBased.__init__(self, net=self.net)
        Trainable.__init__(self)

    def get_checkpoint_filename_suffix(self) -> str:
        return ".pth"

    def get_training_output_dir(self) -> str:
        return self.configuration.train_config.argparse_config.work_dir

    @staticmethod
    def _get_dataset_type() -> str:
        return "MLCVZooMMDetDataset"

    def get_net(self) -> Optional[torch.nn.Module]:
        return self.net

    def __update_mm_config(self, init_for_inference: bool) -> None:
        """
        Backward compatibility handling for self.configuration.mm_config.
        Update relevant attributes from deprecated parameters.

        Args:
            init_for_inference:  Whether to update the config for training or inference

        Returns:
            None
        """
        if init_for_inference:
            if self.configuration.mm_config.config_path is None:
                self.configuration.mm_config.config_path = (
                    self.configuration.inference_config.config_path
                )
        else:
            if self.configuration.mm_config.config_path is None:
                self.configuration.mm_config.config_path = (
                    self.configuration.train_config.argparse_config.config
                )

        if self.configuration.mm_config.cfg_options is None:
            self.configuration.mm_config.cfg_options = (
                self.configuration.train_config.argparse_config.cfg_options
            )

    def _get_mmdeploy_config(
        self, runtime: Optional[str] = None
    ) -> Optional[MMDetectionMMDeployConfig]:
        """Check for the existence and return MMDeploy configuration.

        If no runtime is given, the MMDeploy configuration is selected by the runtime attribute of
        the instance.

        Args:
            runtime (Optional[str]): Runtime to get the MMDeploy configuration for. Defaults to
                None.

        Raises:
            ValueError: If no runtime is provided and the model initialization is not finished yet.
            ValueError: If the MMDeploy configuration is missing.
            ValueError: If the runtime is not supported.

        Returns:
            Optional[MMDetectionMMDeployConfig]: The MMdeploy configuration for the runtime. If
                runtime is 'DEFAULT', None is returned.
        """

        if runtime is None:
            try:
                runtime = self.runtime
            except AttributeError as error:
                raise ValueError(
                    "The 'rumtime' attribute is not initialized, make sure to call with a runtime "
                    "or after initialization."
                ) from error

        if runtime == Runtime.DEFAULT:
            return None

        if runtime == Runtime.ONNXRUNTIME:
            if self.configuration.mmdeploy_onnxruntime_config is None:
                raise ValueError(
                    "The mmdeploy_onnxruntime_config must be provided for runtime 'ONNXRUNTIME'."
                )

            return self.configuration.mmdeploy_onnxruntime_config

        if runtime == Runtime.ONNXRUNTIME_FLOAT16:
            if self.configuration.mmdeploy_onnxruntime_float16_config is None:
                raise ValueError(
                    "The mmdeploy_onnxruntime_float16_config must be provided for runtime "
                    "'ONNXRUNTIME_FLOAT16'."
                )

            return self.configuration.mmdeploy_onnxruntime_float16_config

        if runtime == Runtime.TENSORRT:
            if self.configuration.mmdeploy_tensorrt_config is None:
                raise ValueError(
                    "The mmdeploy_tensorrt_config must be provided for runtime 'TENSORRT'."
                )

            return self.configuration.mmdeploy_tensorrt_config

        raise ValueError(
            f"Getting MMDeploy config is not supported for runtime '{self.runtime}'."
        )

    def __init_dataloader(self) -> None:
        dataset_type = self._get_dataset_type()
        # Update the class-mapping configuration for a dataset of the dataloaders,
        # if it matches the given "dataset_type"
        for data_loader in ["train_dataloader", "val_dataloader", "test_dataloader"]:
            if (
                data_loader in self.cfg
                and self.cfg[data_loader] is not None
                and self.configuration.class_mapping is not None
            ):
                # Check basic dataset
                if self.cfg[data_loader].dataset.type == dataset_type:
                    self.cfg[data_loader].dataset.class_mapping_config = (
                        self.configuration.class_mapping.to_dict()
                    )
                # Check nested dataset, e.g. ConcatDataset or MultiImageMixDataset
                elif (
                    "dataset" in self.cfg[data_loader].dataset
                    and self.cfg[data_loader].dataset.dataset.type == dataset_type
                ):
                    self.cfg[data_loader].dataset.dataset.class_mapping_config = (
                        self.configuration.class_mapping.to_dict()
                    )

    def _init_inference_model(self) -> None:
        """Additional model initialization for inference mode.

        The net and inferencer attributes are set based on the runtime attribute and the ckeckpoint
        is loaded if runtime is 'DEFAULT'.

        Raises:
            RuntimeError: If the model is deployed with MMDeploy and the mmdeploy module can not be
                imported.
            ValueError: If the runtime is not supported.
        """
        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                self.inferencer = DetInferencer(
                    self.cfg,
                    weights=None,
                    device=self.configuration.inference_config.device_string,
                )
                self.net = self.inferencer.model

                if self.configuration.inference_config.checkpoint_path != "":
                    self.restore(
                        checkpoint_path=self.configuration.inference_config.checkpoint_path
                    )
        elif self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            # Return value of _get_mmdeploy_config will not be None
            mmdeploy_config = cast(
                MMDetectionMMDeployConfig, self._get_mmdeploy_config()
            )

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            self.inferencer = MMDeployInferencer(
                model_config=self.cfg,
                mmdeploy_config=mmdeploy_config,
            )
        else:
            raise ValueError(
                f"Initialization for inference is not supported for runtime '{self.runtime}'."
            )

    def _init_training_model(self) -> None:
        """Additional model initialization for training mode.

        Raises:
            RuntimeError: If the learning rate autoscaler is not configured properly.
            ValueError: If the runtime is not supported.
        """
        if self.runtime == Runtime.DEFAULT:
            # load config
            self.cfg.launcher = self.configuration.train_config.argparse_config.launcher
            self.cfg.experiment_name = self.unique_name
            # We always define the work_dir in the configuration file
            self.cfg.work_dir = self.configuration.train_config.argparse_config.work_dir

            # enable automatic-mixed-precision training
            if self.configuration.train_config.argparse_config.amp is True:
                optim_wrapper = self.cfg.optim_wrapper.type
                if optim_wrapper == "AmpOptimWrapper":
                    logger.warning(
                        "AMP training is already enabled in your config.",
                    )
                else:
                    assert optim_wrapper == "OptimWrapper", (
                        "`--amp` is only supported when the optimizer wrapper type is "
                        f"`OptimWrapper` but got {optim_wrapper}."
                    )
                    self.cfg.optim_wrapper.type = "AmpOptimWrapper"
                    self.cfg.optim_wrapper.loss_scale = "dynamic"

            # enable automatically scaling LR
            if self.configuration.train_config.argparse_config.auto_scale_lr:
                if (
                    "auto_scale_lr" in self.cfg
                    and "enable" in self.cfg.auto_scale_lr
                    and "base_batch_size" in self.cfg.auto_scale_lr
                ):
                    self.cfg.auto_scale_lr.enable = True
                else:
                    raise RuntimeError(
                        'Can not find "auto_scale_lr" or '
                        '"auto_scale_lr.enable" or '
                        '"auto_scale_lr.base_batch_size" in your'
                        " configuration file."
                    )
        else:
            raise ValueError(
                f"Initialization for training is not supported for runtime '{self.runtime}'."
            )

    def store(self, checkpoint_path: str) -> None:
        pass

    def restore(self, checkpoint_path: str) -> None:
        if self.net is None:
            raise ValueError(
                "In order to restore a checkpoint, the net attribute has"
                "to be initialized!"
            )

        logger.info(
            "Load model for %s from %s",
            self.unique_name,
            checkpoint_path,
        )

        checkpoint = load_checkpoint(
            self.net,
            checkpoint_path,
            map_location="cpu",
        )

        if "CLASSES" in checkpoint.get("meta", {}):
            self.net.CLASSES = checkpoint["meta"]["CLASSES"]

    def save_reduced_checkpoint(
        self, input_checkpoint_path: str, output_checkpoint_path: str
    ) -> None:
        """
        Saves a reduced version of a stored checkpoint that does not contain optimizer states
        anymore. Therefore, it keeps the weights and meta information of the source checkpoint.

        Args:
            input_checkpoint_path: Path to source checkpoint file
            output_checkpoint_path: Path to where the checkpoint is saved
        """

        runner = MMDetectionModel._build_runner(cfg=self.cfg)

        # Load input checkpoint dict from and extract the metadata for the output checkpoint.
        # Save reduced checkpoint with metadata of full checkpoint to target directory.
        runner.save_checkpoint(
            out_dir=os.path.dirname(output_checkpoint_path),
            filename=os.path.basename(output_checkpoint_path),
            save_optimizer=False,
            save_param_scheduler=False,
            meta=runner.load_checkpoint(filename=input_checkpoint_path)["meta"],
        )

        logger.info(
            "Saved checkpoint from '%s' in a reduced version to '%s'.",
            input_checkpoint_path,
            output_checkpoint_path,
        )

    def train(self) -> None:
        if (
            self.configuration.train_config.multi_gpu_config is None
            or "LOCAL_RANK" in os.environ
        ):
            # Don't start a distributed training if called without a multi GPU config,
            # or run as a worker via torch.distributed as indicated by the presence of a local rank
            logger.info("Training model")
            self._train()
        else:
            logger.info("Training distributed model")
            self._train_multi_gpu()

    def __deploy(self, runtime: str, init_for_inference: bool = True) -> None:
        if runtime == Runtime.DEFAULT:
            return

        if not init_for_inference:
            raise ValueError(
                f"Deploying for training is not supported for runtime '{runtime}'."
            )

        if runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            # Return value of _get_mmdeploy_config will not be None
            mmdeploy_config = cast(
                MMDetectionMMDeployConfig, self._get_mmdeploy_config(runtime=runtime)
            )
        else:
            raise ValueError(f"Deployment is not supported for runtime '{runtime}'.")

        if not all(
            os.path.exists(checkpoint_path)
            for checkpoint_path in mmdeploy_config.checkpoint_paths
        ):
            if MMDeployConverter is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to deploy a model with MMDeploy."
                )

            mmdeploy_converter = MMDeployConverter(
                model_config=self.cfg,
                checkpoint_path=self.configuration.inference_config.checkpoint_path,
                mmdeploy_config=mmdeploy_config,
            )
            mmdeploy_converter.run()

    @staticmethod
    def _register_dataset() -> None:
        """
        Register the custom dataset of the MLCVZoo in the registry of mmcv

        Returns:
            None
        """
        DATASETS.register_module(
            MLCVZooMMDetDataset.__name__, module=MLCVZooMMDetDataset, force=True
        )

    def _train_multi_gpu(
        self,
    ) -> None:
        """
        Run mmdet multi-gpu/distributed training.

        Returns:
            None
        """
        if self.configuration.train_config.multi_gpu_config is None:
            raise RuntimeError("No multi GPU config provided")

        distributed_model_config_path = os.path.join(
            self.get_training_output_dir(),
            "{}-worker.yaml".format(self.configuration.unique_name),
        )
        os.makedirs(os.path.dirname(distributed_model_config_path), exist_ok=True)
        with open(distributed_model_config_path, "w+") as yaml_file:
            self.configuration.to_yaml(
                yaml_package=yaml, dumper_cls=yaml.Dumper, stream=yaml_file
            )

        command = (
            f"{sys.executable} -u "
            "-m torch.distributed.run "
            f"--nproc_per_node={self.configuration.train_config.multi_gpu_config.nproc_per_node} "
            f"--master_port={self.configuration.train_config.multi_gpu_config.master_port} "
            f"--master_addr={self.configuration.train_config.multi_gpu_config.master_address} "
            f"--nnodes={self.configuration.train_config.multi_gpu_config.nnodes} "
            f"--node_rank={self.configuration.train_config.multi_gpu_config.node_rank} "
            f"{inspect.getfile(self.__class__)} "
            f"{distributed_model_config_path} "
        )

        subproc_env = dict(os.environ.copy())
        subproc_env["LOGLEVEL"] = "INFO"

        logger.debug("Run command: %s", command)
        result = subprocess.Popen(
            shlex.split(command),
            env=subproc_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=False,
        )
        Thread(
            target=self.__pipe_to_log, args=[result.stdout, logging.INFO], daemon=True
        ).start()
        result.wait()

        if result.returncode:
            logger.error(
                "Command '%s' returned with exit code %i", command, result.returncode
            )
            raise RuntimeError(
                f"Distributed training exited with exitcode != 0, "
                f"exitcode: {result.returncode}"
            )

    @staticmethod
    def __pipe_to_log(pipe: IO[bytes], level: int) -> None:
        with pipe:
            for line in iter(pipe.readline, b""):  # b'\n'-separated lines
                logger.log(level, "Subprocess: %r", line.decode("utf-8").strip())

    @staticmethod
    def _build_runner(cfg: Config) -> Runner:
        from mmengine.registry import RUNNERS

        # build the runner from config
        if "runner_type" not in cfg:
            # build the default runner
            runner = Runner.from_cfg(cfg)
        else:
            # build customized runner from the registry
            # if 'runner_type' is set in the cfg
            runner = RUNNERS.build(cfg)

        return runner

    def _train(
        self,
    ) -> None:
        # Reduce the number of repeated compilations and improve
        # training speed.
        setup_cache_size_limit_of_dynamo()

        runner = MMDetectionModel._build_runner(cfg=self.cfg)

        # start training
        runner.train()

    @staticmethod
    def run(model_class: Callable) -> None:  # type: ignore[type-arg]
        """
        Run training with given model subclass, reading a config from file.
        This function is intended to be called during torch.distributed.run via subclass main
        functions. It expects a single command line parameter called "yaml_config_path" pointing
        to a valid YAML configuration file for the given model subclass.

        Args:
            model_class: the subclass of model.MMDetectionModel to train
        """
        parser: argparse.ArgumentParser = argparse.ArgumentParser()
        parser.add_argument(
            "yaml_config_path",
            help="Load all parameters from the given yaml configuration file",
            type=str,
        )

        args = parser.parse_args()

        mmdet_model = model_class(
            from_yaml=args.yaml_config_path,
            init_for_inference=False,
        )
        mmdet_model.train()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Wrapper to convert MMDetection models using MMDeploy."""

import copy
import os
import shutil
from typing import List, cast

from mmdeploy.apis import torch2onnx
from mmdeploy.apis.utils import to_backend
from mmdeploy.backend.sdk.export_info import export2SDK
from mmdeploy.utils import get_backend, get_ir_config
from mmengine.config import Config

from mlcvzoo_mmdetection.configuration import (
    MMDetectionMMDeployConfig,
    MMDetectionMMDeployOnnxruntimeConfig,
)
from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.structs import MMDeployBackendType
from mlcvzoo_mmdetection.utils import init_mm_config


class MMDeployConverter:  # pylint: disable=too-few-public-methods
    """Converter for MMDetection models using MMDeploy."""

    def __init__(
        self,
        model_config: Config,
        checkpoint_path: str,
        mmdeploy_config: MMDetectionMMDeployConfig,
    ) -> None:
        self.model_config = model_config
        self.checkpoint_path = checkpoint_path
        self.mmdeploy_config = mmdeploy_config
        self.mmdeploy_cfg: Config = init_mm_config(mm_config=mmdeploy_config)

    def run(self) -> None:
        """Run a deployment.

        Raises:
            KeyError: If configuration key 'backend_config' or 'backend_config.type' is missing in
                the MMDeploy configuration.
            NotImplementedError: If the value for key 'backend_config.type' in the MMDeploy
                configuration is not supported.

        Returns:
            None
        """

        try:
            backend_type = self.mmdeploy_cfg["backend_config"]["type"]
        except KeyError as exc:
            raise KeyError(
                "Invalid MMDeploy configuration. Key 'backend_config.type' is not specified."
            ) from exc

        if backend_type == MMDeployBackendType.ONNXRUNTIME:
            return self.__deploy_onnxruntime()

        if backend_type == MMDeployBackendType.TENSORRT:
            return self.__deploy_tensorrt()

        raise NotImplementedError(
            f"Invalid MMDeploy configuration. The backend type '{backend_type}' is not supported."
        )

    def __deploy_onnxruntime(self) -> None:
        """Run a deployment for backend type onnxruntime.

        Raises:
            ValueError: If the value for 'device_string' in the MMDeploy configuration is not
                'cpu'.
        """

        # Narrow type
        mmdeploy_config = cast(
            MMDetectionMMDeployOnnxruntimeConfig, self.mmdeploy_config
        )

        if self.mmdeploy_config.device_string != "cpu":
            raise ValueError("Backend onnxruntime is only supported on CPU.")

        checkpoint_paths = self.__deploy_common()

        # For the onnxruntime backend only one is file generated
        # Move generated end2end.onnx to the correct location
        shutil.move(src=checkpoint_paths[0], dst=mmdeploy_config.checkpoint_path)

    def __deploy_tensorrt(self) -> None:
        """Run a deployment for backend type tensorrt.

        Raises:
            ValueError: If the value for 'device_string' in the MMDeploy configuration does not
                start with 'cuda'.
        """

        # Narrow type
        mmdeploy_config = cast(
            MMDetectionMMDeployOnnxruntimeConfig, self.mmdeploy_config
        )

        if not self.mmdeploy_config.device_string.startswith("cuda"):
            raise ValueError("Backend tensorrt is only supported on GPU.")

        checkpoint_paths = self.__deploy_common()

        # For the tensorrt backend only one file is generated
        # Move generated end2end.engine to the correct location
        shutil.move(src=checkpoint_paths[0], dst=mmdeploy_config.checkpoint_path)

    def __deploy_common(self) -> List[str]:
        """Run common steps for a MMdeploy deployment.

        Returns:
            List[str]: A list of generated backend specific model filepaths.
        """

        if self.mmdeploy_config.dump_info:
            export2SDK(
                deploy_cfg=self.mmdeploy_cfg,
                # Use a deep copy to prevent changes in the model config
                model_cfg=copy.deepcopy(self.model_config),
                work_dir=self.mmdeploy_config.work_dir,
                pth=self.checkpoint_path,
                device=self.mmdeploy_config.device_string,
            )

        # Convert to IR
        ir_config = get_ir_config(self.mmdeploy_cfg)
        ir_save_file = ir_config["save_file"]

        torch2onnx(
            img=self.mmdeploy_config.test_image_path,
            work_dir=self.mmdeploy_config.work_dir,
            save_file=ir_save_file,
            deploy_cfg=self.mmdeploy_cfg,
            model_cfg=self.model_config,
            model_checkpoint=self.checkpoint_path,
            device=self.mmdeploy_config.device_string,
        )

        # Convert to backend
        backend = get_backend(self.mmdeploy_cfg)

        # MMdeploy always returns a list of backend files
        checkpoint_paths: List[str] = to_backend(
            backend_name=backend,
            ir_files=[os.path.join(self.mmdeploy_config.work_dir, ir_save_file)],
            work_dir=self.mmdeploy_config.work_dir,
            deploy_cfg=self.mmdeploy_cfg,
            device=self.mmdeploy_config.device_string,
        )

        return checkpoint_paths



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Wrapper to run converted MMDetection models."""

from typing import Any, List

import torch
from mmdeploy.apis.utils import build_task_processor
from mmdeploy.utils import get_input_shape
from mmdet.apis.det_inferencer import InputsType
from mmengine.config import Config

from mlcvzoo_mmdetection.configuration import MMDetectionMMDeployConfig
from mlcvzoo_mmdetection.utils import init_mm_config


class MMDeployInferencer:  # pylint: disable=too-few-public-methods
    """Inferencer for MMDetection models deployed with MMDeploy."""

    def __init__(
        self,
        model_config: Config,
        mmdeploy_config: MMDetectionMMDeployConfig,
    ) -> None:
        mmdeploy_cfg: Config = init_mm_config(mm_config=mmdeploy_config)

        self.task_processor = build_task_processor(
            model_cfg=model_config,
            deploy_cfg=mmdeploy_cfg,
            device=mmdeploy_config.device_string,
        )

        self.mmdeploy_model = self.task_processor.build_backend_model(
            model_files=mmdeploy_config.checkpoint_paths,
            data_preprocessor_updater=self.task_processor.update_data_preprocessor,
        )

        self._input_shape = get_input_shape(deploy_cfg=mmdeploy_cfg)

    # The return type can be a list of different MMLab data types, on which we are not dependent in
    # this repo.
    def __call__(self, inputs: InputsType) -> List[Any]:
        """Run inference for given inputs.

        To be compatible with MMDetections DetInferencer, the __call__ method is used.

        Args:
            inputs (InputsType): The inputs to run an inference on.

        Returns:
            List[Any]: List of predictions.
        """
        model_inputs, _ = self.task_processor.create_input(
            imgs=inputs, input_shape=self._input_shape
        )

        with torch.no_grad():
            result: List[Any] = self.mmdeploy_model.test_step(model_inputs)

        return result



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Model that wraps all objection detection models of mmdetection
"""
import logging
import typing
from typing import Dict, List, Optional, Tuple, Union, cast

import numpy as np
import torch.nn
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.api.model import ObjectDetectionModel
from mlcvzoo_base.api.structs import Runtime
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from mmdet.apis.det_inferencer import DetInferencer
from mmdet.structures.det_data_sample import DetDataSample

from mlcvzoo_mmdetection.configuration import (
    MMDetectionConfig,
    MMDetectionInferenceConfig,
)
from mlcvzoo_mmdetection.model import ImageType, MMDetectionModel

try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.inferencer import MMDeployInferencer
except ModuleNotFoundError as error:
    MMDeployInferencer = None  # type: ignore # pylint: disable=invalid-name

logger = logging.getLogger(__name__)


class MMObjectDetectionModel(
    MMDetectionModel[MMDetectionInferenceConfig],
    ObjectDetectionModel[MMDetectionConfig, Union[str, ImageType]],
    NetBased[torch.nn.Module, MMDetectionInferenceConfig],
    Trainable,
):
    """Wrapper for MMDetection object detecion models."""

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
        is_multi_gpu_instance: bool = False,
        runtime: str = Runtime.DEFAULT,
    ) -> None:
        MMDetectionModel.__init__(
            self,
            configuration=self.create_configuration(
                from_yaml, configuration, string_replacement_map
            ),
            string_replacement_map=string_replacement_map,
            init_for_inference=init_for_inference,
            is_multi_gpu_instance=is_multi_gpu_instance,
            runtime=runtime,
        )
        ObjectDetectionModel.__init__(
            self,
            configuration=self.configuration,
            mapper=AnnotationClassMapper(
                class_mapping=self.configuration.class_mapping,
                reduction_mapping=self.configuration.inference_config.reduction_class_mapping,
            ),
            init_for_inference=init_for_inference,
            runtime=runtime,
        )
        NetBased.__init__(self, net=self.net)
        Trainable.__init__(self)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> MMDetectionConfig:
        return typing.cast(
            MMDetectionConfig,
            create_basis_configuration(
                configuration_class=MMDetectionConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @property
    def num_classes(self) -> int:
        return self.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.mapper.annotation_class_id_to_model_class_name_map

    def __decode_mmdet_result(self, prediction: DetDataSample) -> List[BoundingBox]:
        """
        Decode output of an object detection model from mmdetection

        Args:
            prediction: The result that the model has predicted

        Returns:
            The decoded prediction as list of bounding boxes in MLCVZoo format
        """

        bounding_boxes: List[BoundingBox] = list()
        valid_indices = (
            prediction.pred_instances.scores
            > self.configuration.inference_config.score_threshold
        )

        # Filter results according to the determined valid indices
        valid_bounding_boxes = prediction.pred_instances.bboxes[valid_indices]
        valid_class_ids = prediction.pred_instances.labels[valid_indices]
        valid_scores = prediction.pred_instances.scores[valid_indices]

        for bbox, class_id, score in zip(
            valid_bounding_boxes, valid_class_ids, valid_scores
        ):
            bbox_int = bbox.cpu().numpy().astype(np.int_)
            class_id_int = int(class_id.cpu())

            try:
                bounding_boxes.extend(
                    self.build_bounding_boxes(
                        box_format=ObjectDetectionBBoxFormats.XYXY,
                        box_list=(bbox_int[0:4]),
                        class_identifiers=self.mapper.map_model_class_id_to_output_class_identifier(
                            class_id=class_id_int
                        ),
                        model_class_identifier=ClassIdentifier(
                            class_id=class_id_int,
                            class_name=self.mapper.map_annotation_class_id_to_model_class_name(
                                class_id=class_id_int
                            ),
                        ),
                        score=float(score.cpu()),
                        difficult=False,
                        occluded=False,
                        content="",
                    )
                )
            except ValueError as value_error:
                logger.exception(
                    msg="Could not decode bounding box '%s'. Bounding Box will be skipped."
                    % (bbox_int[0:4]),
                    exc_info=value_error,
                )

        return bounding_boxes

    decode_mmdet_result = __decode_mmdet_result

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[BoundingBox]]:
        """Run a prediction on a single input.

        Args:
            data_item (Union[str, ImageType]): The input to run the inference on.

        Raises:
            ValueError: If the net attribute is not initialized and the runtime is 'DEFAULT'.
            ValueError: If the inferencer aatribute is not initialized.
            RuntimeError: If the model is deployed with MMDeploy and the mmdeploy module can not be
                imported.
            ValueError: If the runtime is not supported.

        Returns:
            Tuple[Union[str, ImageType], List[BoundingBox]]: The input and the predicted bounding
                boxes.
        """

        no_inferencer_error = ValueError(
            "The 'inferencer' attribute is not initialized, make sure to instantiate with "
            "init_for_inference=True"
        )

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )
            if self.inferencer is None:
                raise no_inferencer_error

            # For a single data_item we only have one prediction
            return data_item, self.__decode_mmdet_result(
                prediction=cast(DetInferencer, self.inferencer)(
                    inputs=data_item, return_datasample=True, batch_size=1
                )["predictions"][0]
            )
        if self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.inferencer is None:
                raise no_inferencer_error

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            return data_item, self.__decode_mmdet_result(
                prediction=cast(MMDeployInferencer, self.inferencer)(inputs=data_item)[
                    0
                ]
            )

        raise ValueError(f"Prediction is not supported for runtime '{self.runtime}'.")

    def predict_many(
        self, data_items: List[Union[str, ImageType]]
    ) -> List[Tuple[Union[str, ImageType], List[BoundingBox]]]:
        """Run a prediction on a batch of input.

        Args:
            data_items (List[Union[str, ImageType]]): The inputs to run the inference on.

        Raises:
            ValueError: If the net attribute is not initialized and the runtime is 'DEFAULT'.
            ValueError: If the inferencer aatribute is not initialized.
            RuntimeError: If the model is deployed with MMDeploy and the mmdeploy module can not be
                imported.
            ValueError: If the runtime is not supported.

        Returns:
            List[Tuple[Union[str, ImageType], List[BoundingBox]]]: A list of inputs and the
                predicted bounding boxes.
        """

        no_inferencer_error = ValueError(
            "The 'inferencer' attribute is not initialized, make sure to instantiate with "
            "init_for_inference=True"
        )

        prediction_list: List[Tuple[Union[str, ImageType], List[BoundingBox]]] = []

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )
            if self.inferencer is None:
                raise no_inferencer_error

            predictions: List[DetDataSample] = cast(DetInferencer, self.inferencer)(
                inputs=data_items,
                return_datasample=True,
                batch_size=len(data_items),
            )["predictions"]
        elif self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.inferencer is None:
                raise no_inferencer_error

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            predictions = [
                cast(MMDeployInferencer, self.inferencer)(inputs=data_item)[0]
                for data_item in data_items
            ]
        else:
            raise ValueError(
                f"Multi-prediction is not supported for runtime '{self.runtime}'."
            )

        for data_item, prediction in zip(data_items, predictions):
            bounding_boxes = self.__decode_mmdet_result(prediction=prediction)

            prediction_list.append(
                (
                    data_item,
                    bounding_boxes,
                )
            )

        return prediction_list


if __name__ == "__main__":
    MMDetectionModel.run(MMObjectDetectionModel)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition of the MMDetectionConfig that is used to configure the MMDetectionModel.
"""

import logging
from typing import Any, Dict, List, Optional

import related
from attr import Factory, define
from config_builder import BaseConfigClass
from mlcvzoo_base.api.configuration import InferenceConfig, ModelConfiguration
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig

logger = logging.getLogger(__name__)


@define
class MMDetectionModelOverwriteConfig(BaseConfigClass):
    __related_strict__ = True
    num_classes: int = related.IntegerField()

    def check_values(self) -> bool:
        return self.num_classes >= 1


@define
class MMDetectionTrainArgparseConfig(BaseConfigClass):
    __related_strict__ = True
    # Gathering of all argparse parameter mmdetection/tools/train.py:

    # the dir to save logs and models
    work_dir: str = related.StringField()
    # train config file path
    config: Optional[str] = related.ChildField(cls=str, default=None, required=False)
    # enable automatic-mixed-precision training
    amp: bool = related.BooleanField(default=False)
    # enable automatically scaling LR
    auto_scale_lr: bool = related.BooleanField(required=False, default=False)
    # override some settings in the used config, the key-value pair '
    # 'in xxx=yyy format will be merged into config file. If the value to '
    # 'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
    # 'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
    # 'Note that the quotation marks are necessary and that no white space '
    # 'is allowed.
    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )
    # job launcher
    launcher: str = related.StringField(default="none", required=False)

    # ======================================================================================
    # Parameters of the mmdetection/tools/train.py that are not used in mlcvzoo-mmdetection:
    #
    # If specify checkpoint path, resume from it, while if not
    # specify, try to auto resume from the latest checkpoint
    # in the work directory:
    # resume: str
    #
    # NOTE: The local_rank parameter will be set via the argparse argument
    #       in the training scripts.
    # When using PyTorch version >= 2.0.0, the `torch.distributed.launch`
    # will pass the `--local-rank` parameter to `tools/train.py` instead
    # of `--local_rank`:
    #
    # local_rank: int

    def check_values(self) -> bool:
        if self.config is not None:
            logger.warning(
                "configuration.train_config.argparse_config.config is deprecated and will be "
                "removed in future versions."
            )
        if self.cfg_options is not None:
            logger.warning(
                "configuration.train_config.argparse_config.cfg_options is deprecated and will be "
                "removed in future versions."
            )

        return self.launcher in ["none", "pytorch", "slurm", "mpi"]


@define
class MMDetectionDistributedTrainConfig(BaseConfigClass):
    """
    Definition of attributes needed to start a distributed training on multiple GPUs.
    It is leaned against the documentation of mmdetection:
    https://mmdetection.readthedocs.io/en/dev-3.x/user_guides/train.html?highlight=gpu


    TODO: Add this to the arc42 Documentation
    Launch multiple jobs simultaneously
    If you would like to launch multiple jobs on a single machine, e.g., 2 jobs of 4-GPU training
    on a machine with 8 GPUs, you need to specify different ports (29500 by default) for each job
    to avoid communication conflict.

    If you use dist_train.sh to launch training jobs, you can set the port in the commands.

    CUDA_VISIBLE_DEVICES=0,1,2,3 PORT=29500 ./tools/dist_train.sh ${CONFIG_FILE} 4
    CUDA_VISIBLE_DEVICES=4,5,6,7 PORT=29501 ./tools/dist_train.sh ${CONFIG_FILE} 4
    """

    __related_strict__ = True

    # Number of nodes, or the range of nodes in form <minimum_nodes>:<maximum_nodes>
    nnodes: str = related.StringField(default="1:1")

    # Number of workers per node; supported values: [auto, cpu, gpu, int]
    nproc_per_node = related.StringField(default="auto")

    # Rank of the node for multi-node distributed training.
    node_rank: int = related.IntegerField(default=0)

    # Synchronisation port for interprocess communication
    master_port: int = related.IntegerField(default=29500)

    # Synchronisation URI for interprocess communication
    master_address: str = related.StringField(default="localhost")


@define
class MMDetectionTrainConfig(BaseConfigClass):
    """
    argparse parameter from mmdetection/tools/train.py
    """

    __related_strict__ = True

    argparse_config: MMDetectionTrainArgparseConfig = related.ChildField(
        cls=MMDetectionTrainArgparseConfig
    )

    multi_gpu_config: Optional[MMDetectionDistributedTrainConfig] = related.ChildField(
        cls=MMDetectionDistributedTrainConfig,
        required=False,
    )


@define
class MMDetectionInferenceConfig(InferenceConfig):
    __related_strict__ = True

    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )

    device_string: Optional[str] = related.StringField(required=False, default=None)

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )

    def check_values(self) -> bool:
        if self.config_path is None:
            logger.warning(
                "configuration.inference_config.config_path is deprecated and will be removed in "
                "future versions."
            )

        return 0.0 <= self.score_threshold <= 1.0


@define
class MMConfig(BaseConfigClass):
    __related_strict__ = True

    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )

    config_dict: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )

    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )


@define
class MMDetectionMMDeployConfig(MMConfig):
    """Definition of attributes needed to deploy a model using MMDeploy."""

    __related_strict__ = True

    device_string: str = related.StringField()
    work_dir: str = related.StringField()
    test_image_path: str = related.StringField()
    dump_info: bool = related.BooleanField()
    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )
    config_dict: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )
    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )

    @property
    def checkpoint_paths(self) -> List[str]:
        """List of all MMDeploy related checkpoint paths that are needed to run the deployed model.

        Raises:
            NotImplementedError: _description_

        Returns:
            List[str]: List of checkpoint paths.
        """
        raise NotImplementedError(
            "Must be implemented by sub-class: checkpoint_paths(...)."
        )


@define
class MMDetectionMMDeployOnnxruntimeConfig(MMDetectionMMDeployConfig):
    """Definition of attributes needed to deploy a model using MMDeploy with ONNX Runtime."""

    __related_strict__ = True

    checkpoint_path: str = related.StringField()
    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )
    config_dict: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )
    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )

    @property
    def checkpoint_paths(self) -> List[str]:
        return [self.checkpoint_path]


@define
class MMDetectionMMDeployTensorRTConfig(MMDetectionMMDeployConfig):
    """Definition of attributes needed to deploy a model using MMDeploy with TensorRT."""

    __related_strict__ = True

    checkpoint_path: str = related.StringField()
    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )
    config_dict: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )
    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )

    @property
    def checkpoint_paths(self) -> List[str]:
        return [self.checkpoint_path]


@define
class MMDetectionConfig(ModelConfiguration):
    """Definition of attributes for a MMDetectionModel."""

    __related_strict__ = True

    class_mapping: ClassMappingConfig = related.ChildField(cls=ClassMappingConfig)

    inference_config: MMDetectionInferenceConfig = related.ChildField(
        cls=MMDetectionInferenceConfig
    )

    train_config: MMDetectionTrainConfig = related.ChildField(
        cls=MMDetectionTrainConfig
    )

    mmdeploy_onnxruntime_config: Optional[MMDetectionMMDeployOnnxruntimeConfig] = (
        related.ChildField(
            cls=MMDetectionMMDeployOnnxruntimeConfig, default=None, required=False
        )
    )

    mmdeploy_onnxruntime_float16_config: Optional[
        MMDetectionMMDeployOnnxruntimeConfig
    ] = related.ChildField(
        cls=MMDetectionMMDeployOnnxruntimeConfig, default=None, required=False
    )

    mmdeploy_tensorrt_config: Optional[MMDetectionMMDeployTensorRTConfig] = (
        related.ChildField(
            cls=MMDetectionMMDeployTensorRTConfig, default=None, required=False
        )
    )

    mm_config: MMConfig = related.ChildField(
        cls=MMConfig, default=Factory(MMConfig), required=False
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for handling utility methods that are used across the
mlcvzoo_mmdetection package.
"""

import logging
from typing import Any, Dict, Optional

from mmengine.config import Config

from mlcvzoo_mmdetection.configuration import MMConfig

logger = logging.getLogger(__name__)


def init_mmdetection_config(
    config_path: str,
    cfg_options: Optional[Dict[str, Any]] = None,
) -> Config:
    """
    Note: This function is deprecated and will be removed in future versions. Please use
    'mlcvzoo_mmdetection.utils.init_mm_config' instead.

    Initialize an mmengine.config.Config object from a given mmdetection .py config file.

    Args:
        config_path: Filepath of the mmdetection .py config file.
        cfg_options: Overwrite / additional options to load into the mmengine Config object.

    Returns:
        The mmengine Config object.
    """
    logger.warning(
        "DeprecationWarning: 'mlcvzoo_mmdetection.utils.init_mmdetection_config' "
        "is deprecated and will be removed in future versions. Please use "
        "'mlcvzoo_mmdetection.utils.init_mm_config' instead."
    )

    return init_mm_config(
        mm_config=MMConfig(config_path=config_path, cfg_options=cfg_options)
    )


def init_mm_config(mm_config: MMConfig) -> Config:
    """
    Initialize an mmengine.config.Config object from a given
    mmlcvzoo_mmdetection.configuration.MMConfig object. Currently,
    the object can be initialized from its mmdetection .py config
    file as well as from a python dictionary. Please note that
    initialization from .py config files will be removed in future
    releases (cf. documentation for 'mlcvzoo_mmdetection.utils.init_mmdetection_config').

    Args:
        mm_config: The MMConfig object.

    Returns:
        The mmengine Config object.
    """
    if mm_config.config_path is not None:
        # Build config from .py configuration file
        logger.info(
            "Load mmdetection config from config-path: %s", mm_config.config_path
        )
        cfg = Config.fromfile(mm_config.config_path)
    elif mm_config.config_dict is not None:
        logger.info(
            "Load mmdetection config from config-dict: %s", mm_config.config_dict
        )
        cfg = Config(cfg_dict=mm_config.config_dict)
    else:
        raise ValueError(
            "Can not build mmengine config, either config-path or config-dict have to be"
            "provided in mm_config"
        )

    if mm_config.cfg_options:
        cfg.merge_from_dict(mm_config.cfg_options)

    return cfg


def modify_config(config_path: str, string_replacement_map: Dict[str, str]) -> str:
    """
    Note: This function is deprecated and will be removed in future versions.

    Load a mmdetection .py config file via its filepath and replace all matching keys
    (= placeholder / environment variables) in the file with their corresponding values
    in the string_replacement_map. The replaced file content is then stored in a *_local.py
    file in the same directory.

    Args:
        config_path: Path to the .py config file.
        string_replacement_map: Dictionary with string replacement values, e.g.
                                {"PROJECT_ROOT_DIR": "your/local/project/dir"}

    Returns:
        The filepath of the modified *_local.py file.
    """
    with open(file=config_path, mode="r", encoding="'utf-8") as config_file:
        config_file_content = config_file.readlines()

    new_config_file_content = list()
    for config_content in config_file_content:
        new_config_content = config_content

        for replacement_key, replacement_value in string_replacement_map.items():
            if replacement_key in config_content:
                new_config_content = new_config_content.replace(
                    replacement_key, replacement_value
                )

                logger.info(
                    "Replace '%s' in config-line '%s' with '%s'",
                    replacement_key,
                    new_config_content,
                    replacement_value,
                )

        new_config_file_content.append(new_config_content)

    new_config_path = config_path.replace(".py", "_local.py")
    with open(file=new_config_path, mode="w+", encoding="'utf-8") as new_config_file:
        new_config_file.writelines(new_config_file_content)

    return new_config_path



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for defining MMDeploy related structs."""

from enum import Enum


class MMDeployBackendType(str, Enum):
    """Defines backend types used in MMDeploy."""

    ONNXRUNTIME = "onnxruntime"
    TENSORRT = "tensorrt"
